import os
import toga
from toga.style import Pack 
from toga.style.pack import COLUMN, ROW
from cwl_airflow.app.launch import Launcher


class App(toga.App):

    __airflow_home_default = "~/airflow"
    __launcher = None

    def load(self, widget):
        self.__launcher.load()

    def unload(self, widget):
        self.__launcher.unload()

    def startup(self):
        # Create Launcher
        self.__launcher = Launcher(self.__airflow_home_default)

        # Create a main window with a name matching the app
        self.main_window = toga.MainWindow(title=self.name, size=(250,100))

        # Create a main content box
        main_box = toga.Box()
        main_box.style.update(direction=ROW, padding=20)

        # Add load button
        load_button = toga.Button("Load", on_press=self.load)
        load_button.style.padding = 25
        main_box.add(load_button)

        # Add unload button
        unload_button = toga.Button("Unload", on_press=self.unload)
        unload_button.style.padding = 25
        main_box.add(unload_button)

        # Add the content on the main window
        self.main_window.content = main_box

        # Show the main window
        self.main_window.show()

def main():
    return App('CWL-Airflow Control Panel', 'com.biowardrobe')



import os
import re
import subprocess
import configparser
import shutil
from plistlib import load, dump
from airflow.utils.db import merge_conn
from airflow import models


class Launcher:

    __AIRFLOW_SCH = "~/Library/LaunchAgents/com.biowardrobe.airflow-scheduler.plist"
    __AIRFLOW_WEB = "~/Library/LaunchAgents/com.biowardrobe.airflow-webserver.plist"
    __AIRFLOW_API = "~/Library/LaunchAgents/com.biowardrobe.airflow-apiserver.plist"

    __airflow_home = None
    __airflow_cfg = None
    __sch_conf = None
    __web_conf = None
    __api_conf = None
    __contents_dir = None


    def __init__(self, airflow_home = "~/airflow"):
        self.__contents_dir = os.path.dirname(os.path.abspath(os.path.join(__file__, "../../../../")))
        self.__airflow_home = os.path.expanduser(airflow_home)
        self.__airflow_cfg = os.path.join(self.__airflow_home, "airflow.cfg")
        self.__configure()


    def __configure(self):
        try:
            self.__sch_conf = self.__read_plist(os.path.expanduser(self.__AIRFLOW_SCH))
            self.__web_conf = self.__read_plist(os.path.expanduser(self.__AIRFLOW_WEB))
            self.__api_conf = self.__read_plist(os.path.expanduser(self.__AIRFLOW_API))
        except:
            default_conf_folder = os.path.join(os.path.dirname(os.path.abspath(os.path.join(__file__, "../"))), "scripts", "macos")
            self.__sch_conf = self.__read_plist(os.path.join(default_conf_folder, os.path.basename(self.__AIRFLOW_SCH)))
            self.__web_conf = self.__read_plist(os.path.join(default_conf_folder, os.path.basename(self.__AIRFLOW_WEB)))
            self.__api_conf = self.__read_plist(os.path.join(default_conf_folder, os.path.basename(self.__AIRFLOW_API)))
            self.__update_variables()
            self.__update_shebang()
            self.__init_airflow_db()
            self.__update_airflow_config()
            self.__copy_dags()
            self.__add_connections()
            self.__write_plist(self.__sch_conf, os.path.expanduser(self.__AIRFLOW_SCH))
            self.__write_plist(self.__web_conf, os.path.expanduser(self.__AIRFLOW_WEB))
            self.__write_plist(self.__api_conf, os.path.expanduser(self.__AIRFLOW_API))


    def __read_plist(self, path):
        with open(path, 'rb') as fs:
            return load(fs)    


    def __write_plist(self, data, path):
        with open(path, 'wb') as fs:
            dump(data, fs)


    def __get_path(self):
        path_list = [
            os.path.join(self.__contents_dir, "Resources/python/bin"),
            os.path.join(self.__contents_dir, "Resources/app_packages/bin"),
            os.path.join(self.__contents_dir, "Resources/app/bin"),
            os.path.join(self.__contents_dir, "MacOS"),
            "/usr/local/bin",
            "/usr/local/sbin",
            "/usr/bin",
            "/bin",
            "/usr/sbin",
            "/sbin"
        ]
        return ":".join(dict.fromkeys(path_list).keys())


    def __get_pythonpath(self):
        pythonpath_list = [
            os.path.join(self.__contents_dir, "Resources/app"),
            os.path.join(self.__contents_dir, "Resources/app_packages")
        ]
        return ":".join(dict.fromkeys(pythonpath_list).keys())


    def __update_variables(self):
        path = self.__get_path()
        pythonpath = self.__get_pythonpath()
        airflow_exe = os.path.join(self.__contents_dir, "Resources/app_packages/bin/airflow")
        cwlairflow_exe = os.path.join(self.__contents_dir, "Resources/app/bin/cwl-airflow")
        log_sch = os.path.join(self.__airflow_home, os.path.splitext(os.path.basename(self.__AIRFLOW_SCH))[0])
        log_web = os.path.join(self.__airflow_home, os.path.splitext(os.path.basename(self.__AIRFLOW_WEB))[0])
        log_api = os.path.join(self.__airflow_home, os.path.splitext(os.path.basename(self.__AIRFLOW_API))[0])

        self.__sch_conf["EnvironmentVariables"]["PATH"] = path
        self.__web_conf["EnvironmentVariables"]["PATH"] = path
        self.__api_conf["EnvironmentVariables"]["PATH"] = path

        self.__sch_conf["EnvironmentVariables"]["PYTHONPATH"] = pythonpath
        self.__web_conf["EnvironmentVariables"]["PYTHONPATH"] = pythonpath
        self.__api_conf["EnvironmentVariables"]["PYTHONPATH"] = pythonpath

        self.__sch_conf["EnvironmentVariables"]["AIRFLOW_HOME"] = self.__airflow_home
        self.__web_conf["EnvironmentVariables"]["AIRFLOW_HOME"] = self.__airflow_home
        self.__api_conf["EnvironmentVariables"]["AIRFLOW_HOME"] = self.__airflow_home

        self.__sch_conf["ProgramArguments"][0] = airflow_exe
        self.__web_conf["ProgramArguments"][0] = airflow_exe
        self.__api_conf["ProgramArguments"][0] = cwlairflow_exe

        self.__sch_conf["WorkingDirectory"] = self.__airflow_home
        self.__web_conf["WorkingDirectory"] = self.__airflow_home
        self.__api_conf["WorkingDirectory"] = self.__airflow_home

        self.__sch_conf["StandardErrorPath"] = log_sch + ".stderr"
        self.__sch_conf["StandardOutPath"] = log_sch + ".stdout"

        self.__web_conf["StandardErrorPath"] = log_web + ".stderr"
        self.__web_conf["StandardOutPath"] = log_web + ".stdout"

        self.__api_conf["StandardErrorPath"] = log_api + ".stderr"
        self.__api_conf["StandardOutPath"] = log_api + ".stdout"


    def __update_shebang(self):
        lookup_dir = os.path.join(self.__contents_dir, "Resources/app_packages/bin")
        shebang = "#!{}".format(os.path.join(self.__contents_dir, "Resources/python/bin/python3"))
        for filename in os.listdir(lookup_dir):
            filename = os.path.join(lookup_dir, filename)
            if not os.path.isfile(filename):
                continue
            with open(filename, "rb") as f:
                try:
                    lines = f.read().decode("utf-8").splitlines()
                except UnicodeDecodeError:
                    continue
            if not lines:
                continue
            script = [shebang] + lines[1:]
            with open(filename, "wb") as f:
                f.write("\n".join(script).encode("utf-8"))


    def __init_airflow_db(self):
        env = {
            "AIRFLOW_HOME": self.__airflow_home,
            "PATH": self.__get_path(),
            "PYTHONPATH": self.__get_pythonpath()
        }
        subprocess.run(["airflow", "initdb"], env=env)


    def __get_configuration(self):
        conf = configparser.ConfigParser()
        conf.read(self.__airflow_cfg)
        return conf


    def __update_airflow_config(self):
        conf = self.__get_configuration()
        with open(self.__airflow_cfg, 'w') as f:
            try:
                conf.add_section('cwl')
            except configparser.DuplicateSectionError:
                pass
            conf.set("cwl", "tmp_folder", os.path.join(self.__airflow_home, 'tmp'))
            conf.set("core", "logging_level", "INFO")
            conf.set("core", "load_examples", "False")
            conf.set("core", "dags_are_paused_at_creation", "False")
            conf.set("webserver", "dag_default_view", "graph")
            conf.set("webserver", "dag_orientation", "TB")
            conf.set("webserver", "hide_paused_dags_by_default", "True")
            conf.write(f)


    def __copy_dags(self):
        conf = self.__get_configuration()
        source_dags_folder = os.path.join(os.path.dirname(os.path.abspath(os.path.join(__file__, "../"))), "dags")
        target_dags_folder = conf.get("core", "dags_folder")
        os.makedirs(target_dags_folder, mode=0o0775, exist_ok=True)
        for root, dirs, files in os.walk(source_dags_folder):
            for filename in files:
                if re.match(".*\.py$", filename) and filename != "__init__.py":
                    shutil.copy(os.path.join(root, filename), target_dags_folder)


    def __add_connections(self):
        merge_conn(models.Connection(conn_id = "process_report",
                                     conn_type = "http",
                                     host = "localhost",
                                     port = "3070",
                                     extra = "{\"endpoint\":\"/airflow/\"}"))


    def load(self):
        subprocess.run(["launchctl", "load", "-w", os.path.expanduser(self.__AIRFLOW_SCH)])
        subprocess.run(["launchctl", "load", "-w", os.path.expanduser(self.__AIRFLOW_WEB)])
        subprocess.run(["launchctl", "load", "-w", os.path.expanduser(self.__AIRFLOW_API)])


    def unload(self):
        subprocess.run(["launchctl", "unload", os.path.expanduser(self.__AIRFLOW_SCH)])
        subprocess.run(["launchctl", "unload", os.path.expanduser(self.__AIRFLOW_WEB)])
        subprocess.run(["launchctl", "unload", os.path.expanduser(self.__AIRFLOW_API)])


import logging
import os
import shutil
from json import dumps
from jsonmerge import merge

from airflow.models import BaseOperator
from airflow.utils import apply_defaults

from cwltool.process import relocateOutputs
from cwltool.stdfsaccess import StdFsAccess
from cwl_airflow.utils.notifier import task_on_success, task_on_failure, task_on_retry, post_status
from ..cwlstepoperator import CWLStepOperator

_logger = logging.getLogger(__name__)


class CWLJobGatherer(BaseOperator):

    ui_color = '#1E88E5'
    ui_fgcolor = '#FFF'

    @apply_defaults
    def __init__(
            self,
            task_id=None,
            reader_task_id=None,
            *args, **kwargs):
        task_id = task_id if task_id else self.__class__.__name__

        kwargs.update({"on_success_callback": kwargs.get("on_success_callback", task_on_success),
                       "on_failure_callback": kwargs.get("on_failure_callback", task_on_failure),
                       "on_retry_callback":   kwargs.get("on_retry_callback",   task_on_retry)})

        super(CWLJobGatherer, self).__init__(task_id=task_id, *args, **kwargs)

        self.outputs = self.dag.get_output_list()
        self.outdir = None
        self.output_folder = None
        self.reader_task_id = None
        self.reader_task_id = reader_task_id if reader_task_id else self.reader_task_id

    def cwl_gather(self, context):
        upstream_task_ids = [t.task_id for t in self.dag.tasks if isinstance(t, CWLStepOperator)] + \
                            ([self.reader_task_id] if self.reader_task_id else [])
        upstream_data = self.xcom_pull(context=context, task_ids=upstream_task_ids)

        _logger.debug('{0}: xcom_pull data: \n {1}'.
                      format(self.task_id, dumps(upstream_data, indent=4)))

        promises = {}
        for data in upstream_data:
            promises = merge(promises, data["promises"])
            if "outdir" in data:
                self.outdir = data["outdir"]

        if "output_folder" in promises:
            self.output_folder = os.path.abspath(promises["output_folder"])
        else:
            return

        _move_job = {out: promises[out]
                     for out, val in self.outputs.items()
                     }
        _logger.debug('{0}: Final job: \n{1}\nMoving data: \n{2}\nMoving job:{3}'.
                      format(self.task_id,
                             dumps(promises, indent=4),
                             dumps(self.outputs, indent=4),
                             dumps(_move_job, indent=4)))

        _files_moved = relocateOutputs(_move_job, self.output_folder,
                                       [self.outdir], self.dag.default_args["move_outputs"],
                                       StdFsAccess(""))
        _job_result = {val.split("/")[-1]: _files_moved[out]  # TODO: is split required?
                       for out, val in self.outputs.items()
                       if out in _files_moved
                       }
        try:
            if self.outdir:
                shutil.rmtree(self.outdir, ignore_errors=False)
            _logger.info('{0}: Delete temporary output directory {1}'.format(self.task_id, self.outdir))
        except Exception as e:
            _logger.error("{0}: Temporary output directory hasn't been set {1}".format(self.task_id, e))
            pass
        _logger.info("Job done: {}".format(dumps(_job_result, indent=4)))

        return _job_result, promises

    def execute(self, context):
        post_status(context)
        return self.cwl_gather(context)



#!/usr/bin/env python
import connexion
from connexion.resolver import Resolver
from cwl_airflow.wes.backend import CWLAirflowBackend


def run_wes_server(args):
    app = connexion.App(__name__)
    backend = CWLAirflowBackend()
    def rs(x):
        return getattr(backend, x.split('.')[-1])
    app.add_api('openapi/swagger_configuration.yaml', resolver=Resolver(rs))
    app.run(port=args.port, host=args.host)


import sys
import jwt
import socketserver
from http.server import SimpleHTTPRequestHandler
from json import dumps, loads

try:
    PORT = int(sys.argv[1])
except Exception:
    PORT = 8080

public_key = b"""-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDdlatRjRjogo3WojgGHFHYLugd
UWAY9iR3fy4arWNA1KoS8kVw33cJibXr8bvwUAUparCwlvdbH6dvEOfou0/gCFQs
HUfQrSDv+MuSUMAe8jzKE4qW+jK+xQU9a03GUnKHkkle+Q0pX/g6jXZ7r1/xAK5D
o2kQ+X5xK9cipRgEKwIDAQAB
-----END PUBLIC KEY-----
"""

private_key = b"""-----BEGIN RSA PRIVATE KEY-----
MIICWwIBAAKBgQDdlatRjRjogo3WojgGHFHYLugdUWAY9iR3fy4arWNA1KoS8kVw
33cJibXr8bvwUAUparCwlvdbH6dvEOfou0/gCFQsHUfQrSDv+MuSUMAe8jzKE4qW
+jK+xQU9a03GUnKHkkle+Q0pX/g6jXZ7r1/xAK5Do2kQ+X5xK9cipRgEKwIDAQAB
AoGAD+onAtVye4ic7VR7V50DF9bOnwRwNXrARcDhq9LWNRrRGElESYYTQ6EbatXS
3MCyjjX2eMhu/aF5YhXBwkppwxg+EOmXeh+MzL7Zh284OuPbkglAaGhV9bb6/5Cp
uGb1esyPbYW+Ty2PC0GSZfIXkXs76jXAu9TOBvD0ybc2YlkCQQDywg2R/7t3Q2OE
2+yo382CLJdrlSLVROWKwb4tb2PjhY4XAwV8d1vy0RenxTB+K5Mu57uVSTHtrMK0
GAtFr833AkEA6avx20OHo61Yela/4k5kQDtjEf1N0LfI+BcWZtxsS3jDM3i1Hp0K
Su5rsCPb8acJo5RO26gGVrfAsDcIXKC+bQJAZZ2XIpsitLyPpuiMOvBbzPavd4gY
6Z8KWrfYzJoI/Q9FuBo6rKwl4BFoToD7WIUS+hpkagwWiz+6zLoX1dbOZwJACmH5
fSSjAkLRi54PKJ8TFUeOP15h9sQzydI8zJU+upvDEKZsZc/UhT/SySDOxQ4G/523
Y0sz/OZtSWcol/UMgQJALesy++GdvoIDLfJX5GBQpuFgFenRiRDabxrE9MNUZ2aP
FaFp+DyAe+b4nDwuJaW2LURbr8AEZga7oQj0uYxcYw==
-----END RSA PRIVATE KEY-----
"""


class Handler(SimpleHTTPRequestHandler):
    def do_POST(self):
        self.send_response(200)
        self.end_headers()
        headers = self.headers
        data = loads(self.rfile.read(int(self.headers['Content-Length'])).decode("UTF-8"))
        try:
            data["payload"] = jwt.decode(data["payload"].encode("utf-8"), public_key, algorithms='RS256')
        except Exception:
            print("\nFailed to check signature of it wasn't necessary\n")
        print("\nHeaders:\n", headers, "\nData\n", dumps(data, indent=4))

print("Listening on PORT", PORT)

httpd = socketserver.TCPServer(("", PORT), Handler)
httpd.serve_forever()



import subprocess
import logging
import connexion
import random
import string

from os import path, environ, makedirs
from airflow.models import DagBag, TaskInstance, DagRun
from airflow.utils.timezone import parse as parsedate
from airflow.api.common.experimental import trigger_dag
from airflow.settings import DAGS_FOLDER


logger = logging.getLogger(__name__)


class CWLAirflowBackend():

    # curl -X GET "127.0.0.1:8081/wes/v1/dags" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dags?dag_ids=example_bash_operator" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dags?dag_ids=example_bash_operator,tutorial" -H "accept: application/json"

    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?dag_id=example_bash_operator&state=running" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?dag_id=example_bash_operator&state=success" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?dag_id=example_bash_operator" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?state=running" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs" -H "accept: application/json"

    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?run_id=scheduled__2019-07-20T00%3A00%3A00%2B00%3A00" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?run_id=scheduled__2019-07-20T00%3A00%3A00%2B00%3A00&dag_id=tutorial" -H "accept: application/json"
    # curl -X GET "127.0.0.1:8081/wes/v1/dag_runs?execution_date=2019-07-20T00%3A00%3A00%2B00%3A00" -H "accept: application/json"

    # curl -X POST "127.0.0.1:8081/wes/v1/dag_runs?dag_id=example_bash_operator&run_id=1234567&conf=%22%7B%7D%22" -H "accept: application/json"
    # curl -X POST "localhost:8081/wes/v1/dags?dag_id=bowtie2-indices" -H "accept: application/json" -H "Content-Type: multipart/form-data" -F "workflow=@bowtie2-indices.cwl"
    # curl -X POST "localhost:8081/wes/v1//dags/bowtie2-indices/dag_runs?run_id=bowtie2_indices_1&conf=%22%7B%7D%22" -H "accept: application/json"

    env = None
    include_examples = False
    dag_template = "#!/usr/bin/env python3\nfrom cwl_airflow import CWLDAG, CWLJobDispatcher, CWLJobGatherer\ndag = CWLDAG(cwl_workflow='{0}', dag_id='{1}')\ndag.create()\ndag.add(CWLJobDispatcher(dag=dag), to='top')\ndag.add(CWLJobGatherer(dag=dag), to='bottom')"


    def __init__(self):
        makedirs(DAGS_FOLDER, mode=0o0775, exist_ok=True)
        extend_env = {"AIRFLOW__CORE__LOGGING_LEVEL": "ERROR"}
        self.env = environ.copy()
        self.env.update(extend_env)


    def get_dags(self, dag_ids=[]):
        logger.debug(f"""Call get_dags with dag_ids={dag_ids}""")
        try:
            dag_ids = dag_ids or self.list_dags()
            logger.debug(f"""Processing dags {dag_ids}""")
            return {"dags": [{"dag_id": dag_id, "tasks": self.list_tasks(dag_id)} for dag_id in dag_ids]}
        except Exception as err:
            logger.error(f"""Failed while running get_dags {err}""")
            return {"dags": []}


    def post_dag(self, dag_id=None):
        logger.debug(f"""Call post_dag with dag_id={dag_id}""")
        try:
            res = self.export_dag(dag_id or ''.join(random.choice(string.ascii_lowercase) for i in range(32)))
            logger.debug(f"""Exported DAG {res}""")
            return res
        except Exception as err:
            logger.error(f"""Failed while running post_dag {err}""")
            return connexion.problem(500, "Failed to create dag", str(err))


    def get_dag_runs(self, dag_id=None, run_id=None, execution_date=None, state=None):
        logger.debug(f"""Call get_dag_runs with dag_id={dag_id}, run_id={run_id}, execution_date={execution_date}, state={state}""")
        try:
            dag_runs = []
            dag_ids = [dag_id] if dag_id else self.list_dags()
            logger.debug(f"""Processing dags {dag_ids}""")
            for d_id in dag_ids:
                logger.debug(f"""Process dag  {d_id}""")
                task_ids = self.list_tasks(d_id)
                logger.debug(f"""Fetched tasks {task_ids}""")
                for dag_run in self.list_dag_runs(d_id, state):
                    logger.debug(f"""Process dag run {dag_run["run_id"]}, {dag_run["execution_date"]}""")
                    if run_id and run_id != dag_run["run_id"] or execution_date and execution_date != dag_run["execution_date"]:
                        logger.debug(f"""Skip dag_run {dag_run["run_id"]}, {dag_run["execution_date"]} (run_id or execution_date doesn't match)""")
                        continue
                    response_item = {"dag_id": d_id,
                                     "run_id": dag_run["run_id"],
                                     "execution_date": dag_run["execution_date"],
                                     "start_date": dag_run["start_date"],
                                     "state": dag_run["state"],
                                     "tasks": []}
                    logger.debug(f"""Get statuses for tasks {task_ids}""")
                    for t_id in task_ids:
                        response_item["tasks"].append({"id": t_id, "state": self.task_state(d_id, t_id, dag_run["execution_date"])})
                    dag_runs.append(response_item)
            return {"dag_runs": dag_runs}
        except Exception as err:
            logger.error(f"""Failed to call get_dag_runs {err}""")
            return {"dag_runs": []}
        

    def post_dag_runs(self, dag_id, run_id=None, conf=None):
        logger.debug(f"""Call post_dag_runs with dag_id={dag_id}, run_id={run_id}, conf={conf}""")
        try:
            dagrun = self.trigger_dag(dag_id, run_id, conf)
            return {"dag_id": dagrun.dag_id,
                    "run_id": dagrun.run_id,
                    "execution_date": dagrun.execution_date,
                    "start_date": dagrun.start_date,
                    "state": dagrun.state}
        except Exception as err:
            logger.error(f"""Failed to call post_dag_runs {err}""")
            return connexion.problem(500, "Failed to create dag_run", str(err))


    # delete this function if path /dags/{dag_id}/dag_runs isn't actively used
    def post_dag_runs_legacy(self, dag_id, run_id=None, conf=None):
        """For Swagger 2.0 the id MUST be unique among all operations described in the API"""
        return self.post_dag_runs(dag_id, run_id, conf)


    def trigger_dag(self, dag_id, run_id, conf):
        return trigger_dag.trigger_dag(dag_id=dag_id, run_id=run_id, conf=conf, replace_microseconds=False)


    def list_dags(self):
        return DagBag(include_examples=self.include_examples).dags.keys()


    def list_tasks(self, dag_id):
        return [t.task_id for t in DagBag(include_examples=self.include_examples).dags[dag_id].tasks]


    def task_state(self, dag_id, task_id, execution_date):
        task_state = TaskInstance(DagBag(include_examples=self.include_examples).dags[dag_id].get_task(task_id=task_id), parsedate(execution_date)).current_state()
        task_state = task_state or "none"
        return task_state


    def list_dag_runs(self, dag_id, state):
        dag_runs = []
        for dag_run in DagRun.find(dag_id=dag_id, state=state):
            dag_runs.append({
                'run_id': dag_run.run_id,
                'state': dag_run.state,
                'execution_date': dag_run.execution_date.isoformat(),
                'start_date': ((dag_run.start_date or '') and dag_run.start_date.isoformat())
            })
        return dag_runs


    def save_attachment(self, attachment, location, exist_ok=False):
        if path.isfile(location) and not exist_ok:
            raise FileExistsError("[Errno 17] File exists: '" + location + "'")
        data = connexion.request.files[attachment]
        data.save(location)
    

    def export_dag(self, dag_id):
        cwl_path = path.join(DAGS_FOLDER, dag_id + ".cwl")
        dag_path = path.join(DAGS_FOLDER, dag_id + ".py")
        self.save_attachment("workflow", cwl_path)
        with open(dag_path, 'x') as o_stream:
            o_stream.write(self.dag_template.format(cwl_path, dag_id))
        return {"dag_id": dag_id, "cwl_path": cwl_path, "dag_path": dag_path}


#! /usr/bin/env python3

import logging
import json
import os, sys, tempfile
import copy
import glob
import subprocess
import shutil
from jsonmerge import merge

from cwltool.executors import SingleJobExecutor
from cwltool.stdfsaccess import StdFsAccess
from cwltool.workflow import expression
from cwltool.context import RuntimeContext, getdefault
from cwltool.pathmapper import visit_class
from cwltool.mutation import MutationManager

from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults

from .cwlutils import flatten, shortname, load_cwl
from cwl_airflow.utils.notifier import task_on_success, task_on_failure, task_on_retry, post_status

from airflow.utils.log.logging_mixin import StreamLogWriter

_logger = logging.getLogger(__name__)

class StreamLogWriterUpdated (StreamLogWriter):

    def fileno(self):
        return -1


class CWLStepOperator(BaseOperator):

    ui_color = '#3E53B7'
    ui_fgcolor = '#FFF'

    @apply_defaults
    def __init__(
            self,
            task_id=None,
            reader_task_id=None,
            ui_color=None,
            *args, **kwargs):

        self.outdir = None
        self.reader_task_id = None
        self.cwlwf = None
        self.cwl_step = None

        kwargs.update({"on_success_callback": kwargs.get("on_success_callback", task_on_success),
                       "on_failure_callback": kwargs.get("on_failure_callback", task_on_failure),
                       "on_retry_callback":   kwargs.get("on_retry_callback",   task_on_retry)})

        super(self.__class__, self).__init__(task_id=task_id, *args, **kwargs)

        self.reader_task_id = reader_task_id if reader_task_id else self.reader_task_id

        if ui_color:
            self.ui_color = ui_color

    def execute(self, context):

        post_status(context)

        self.cwlwf, it_is_workflow = load_cwl(self.dag.default_args["cwl_workflow"], self.dag.default_args)
        self.cwl_step = [step for step in self.cwlwf.steps if self.task_id == step.id.split("#")[-1]][0] if it_is_workflow else self.cwlwf

        _logger.info('{0}: Running!'.format(self.task_id))

        upstream_task_ids = [t.task_id for t in self.upstream_list] + \
                            ([self.reader_task_id] if self.reader_task_id else [])
        _logger.debug('{0}: Collecting outputs from: \n{1}'.format(self.task_id,
                                                                   json.dumps(upstream_task_ids, indent=4)))

        upstream_data = self.xcom_pull(context=context, task_ids=upstream_task_ids)
        _logger.info('{0}: Upstream data: \n {1}'.format(self.task_id,
                                                         json.dumps(upstream_data, indent=4)))

        promises = {}
        for data in upstream_data:  # upstream_data is an array with { promises and outdir }
            promises = merge(promises, data["promises"])
            if "outdir" in data:
                self.outdir = data["outdir"]

        _d_args = self.dag.default_args

        if not self.outdir:
            self.outdir = _d_args['tmp_folder']

        _logger.debug('{0}: Step inputs: {1}'.format(self.task_id,
                                                     json.dumps(self.cwl_step.tool["inputs"], indent=4)))

        _logger.debug('{0}: Step outputs: {1}'.format(self.task_id,
                                                      json.dumps(self.cwl_step.tool["outputs"], indent=4)))

        jobobj = {}

        for inp in self.cwl_step.tool["inputs"]:
            jobobj_id = shortname(inp["id"]).split("/")[-1]
            source_ids = []
            promises_outputs = []
            try:
                source_field = inp["source"] if it_is_workflow else inp.get("id")
                source_ids = [shortname(s) for s in source_field] if isinstance(source_field, list) else [shortname(source_field)]
                promises_outputs = [promises[source_id] for source_id in source_ids if source_id in promises]
            except:
                _logger.warning("{0}: Couldn't find source field in step input: {1}"
                                .format(self.task_id,
                                        json.dumps(inp, indent=4)))

            _logger.info('{0}: For input {1} with source_ids: {2} found upstream outputs: \n{3}'
                         .format(self.task_id,
                                 jobobj_id,
                                 source_ids,
                                 promises_outputs))

            if len(promises_outputs) > 1:
                if inp.get("linkMerge", "merge_nested") == "merge_flattened":
                    jobobj[jobobj_id] = flatten(promises_outputs)
                else:
                    jobobj[jobobj_id] = promises_outputs
            # Should also check if [None], because in this case we need to take default value
            elif len(promises_outputs) == 1 and (promises_outputs[0] is not None):
                jobobj[jobobj_id] = promises_outputs[0]
            elif "valueFrom" in inp:
                jobobj[jobobj_id] = None
            elif "default" in inp:
                d = copy.copy(inp["default"])
                jobobj[jobobj_id] = d
            else:
                continue

        _logger.debug('{0}: Collected job object: \n {1}'.format(self.task_id, json.dumps(jobobj, indent=4)))

        def _post_scatter_eval(shortio, cwl_step):
            _value_from = {
                shortname(i["id"]).split("/")[-1]:
                    i["valueFrom"] for i in cwl_step.tool["inputs"] if "valueFrom" in i
                }
            _logger.debug(
                '{0}: Step inputs with valueFrom: \n{1}'.format(self.task_id, json.dumps(_value_from, indent=4)))

            def value_from_func(k, v):
                if k in _value_from:
                    return expression.do_eval(
                        _value_from[k], shortio,
                        self.cwlwf.tool.get("requirements", []),
                        None, None, {}, context=v)
                else:
                    return v
            return {k: value_from_func(k, v) for k, v in shortio.items()}

        job = _post_scatter_eval(jobobj, self.cwl_step)
        _logger.info('{0}: Final job data: \n {1}'.format(self.task_id, json.dumps(job, indent=4)))

        _d_args['outdir'] = tempfile.mkdtemp(prefix=os.path.join(self.outdir, "step_tmp"))
        _d_args['tmpdir_prefix'] = os.path.join(_d_args['outdir'], 'cwl_tmp_')
        _d_args['tmp_outdir_prefix'] = os.path.join(_d_args['outdir'], 'cwl_outdir_')

        _d_args["record_container_id"] = True
        _d_args["cidfile_dir"] = _d_args['outdir']
        _d_args["cidfile_prefix"] = self.task_id

        _logger.debug(
            '{0}: Runtime context: \n {1}'.format(self, _d_args))

        executor = SingleJobExecutor()
        runtimeContext = RuntimeContext(_d_args)
        runtimeContext.make_fs_access = getdefault(runtimeContext.make_fs_access, StdFsAccess)

        for inp in self.cwl_step.tool["inputs"]:
            if inp.get("not_connected"):
                del job[shortname(inp["id"].split("/")[-1])]

        _stderr = sys.stderr
        sys.stderr = sys.__stderr__
        (output, status) = executor(self.cwl_step.embedded_tool if it_is_workflow else self.cwl_step,
                                    job,
                                    runtimeContext,
                                    logger=_logger)
        sys.stderr = _stderr

        if not output and status == "permanentFail":
            raise ValueError

        _logger.debug(
            '{0}: Embedded tool outputs: \n {1}'.format(self.task_id, json.dumps(output, indent=4)))

        promises = {}


        for out in self.cwl_step.tool["outputs"]:

            out_id = shortname(out["id"])
            jobout_id = out_id.split("/")[-1]
            try:
                promises[out_id] = output[jobout_id]
            except:
                continue

        # Unsetting the Generation from final output object
        visit_class(promises, ("File",), MutationManager().unset_generation)

        data = {"promises": promises, "outdir": self.outdir}

        _logger.info(
            '{0}: Output: \n {1}'.format(self.task_id, json.dumps(data, indent=4)))

        return data

    def on_kill(self):
        _logger.info("Stop docker containers")
        for cidfile in glob.glob(os.path.join(self.dag.default_args["cidfile_dir"], self.task_id + "*.cid")):  # make this better, doesn't look good to read from self.dag.default_args
            try:
                with open(cidfile, "r") as inp_stream:
                    _logger.debug(f"""Read container id from {cidfile}""")
                    command = ["docker", "kill", inp_stream.read()]
                    _logger.debug(f"""Call {" ".join(command)}""")
                    p = subprocess.Popen(command, shell=False)
                    try:
                        p.wait(timeout=10)
                    except subprocess.TimeoutExpired:
                        p.kill()
            except Exception as ex:
                _logger.error(f"""Failed to stop docker container with ID from {cidfile}\n {ex}""")

        # _logger.info(f"""Delete temporary output directory {self.outdir}""")
        # try:
        #     shutil.rmtree(self.outdir)
        # except Exception as ex:
        #     _logger.error(f"""Failed to delete temporary output directory {self.outdir}\n {ex}""")


#! /usr/bin/env python3
import sys
import argparse
import requests
import jwt
from future.moves.urllib.parse import urljoin


public_key = b"""-----BEGIN PUBLIC KEY-----
MIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDdlatRjRjogo3WojgGHFHYLugd
UWAY9iR3fy4arWNA1KoS8kVw33cJibXr8bvwUAUparCwlvdbH6dvEOfou0/gCFQs
HUfQrSDv+MuSUMAe8jzKE4qW+jK+xQU9a03GUnKHkkle+Q0pX/g6jXZ7r1/xAK5D
o2kQ+X5xK9cipRgEKwIDAQAB
-----END PUBLIC KEY-----
"""


private_key = b"""-----BEGIN RSA PRIVATE KEY-----
MIICWwIBAAKBgQDdlatRjRjogo3WojgGHFHYLugdUWAY9iR3fy4arWNA1KoS8kVw
33cJibXr8bvwUAUparCwlvdbH6dvEOfou0/gCFQsHUfQrSDv+MuSUMAe8jzKE4qW
+jK+xQU9a03GUnKHkkle+Q0pX/g6jXZ7r1/xAK5Do2kQ+X5xK9cipRgEKwIDAQAB
AoGAD+onAtVye4ic7VR7V50DF9bOnwRwNXrARcDhq9LWNRrRGElESYYTQ6EbatXS
3MCyjjX2eMhu/aF5YhXBwkppwxg+EOmXeh+MzL7Zh284OuPbkglAaGhV9bb6/5Cp
uGb1esyPbYW+Ty2PC0GSZfIXkXs76jXAu9TOBvD0ybc2YlkCQQDywg2R/7t3Q2OE
2+yo382CLJdrlSLVROWKwb4tb2PjhY4XAwV8d1vy0RenxTB+K5Mu57uVSTHtrMK0
GAtFr833AkEA6avx20OHo61Yela/4k5kQDtjEf1N0LfI+BcWZtxsS3jDM3i1Hp0K
Su5rsCPb8acJo5RO26gGVrfAsDcIXKC+bQJAZZ2XIpsitLyPpuiMOvBbzPavd4gY
6Z8KWrfYzJoI/Q9FuBo6rKwl4BFoToD7WIUS+hpkagwWiz+6zLoX1dbOZwJACmH5
fSSjAkLRi54PKJ8TFUeOP15h9sQzydI8zJU+upvDEKZsZc/UhT/SySDOxQ4G/523
Y0sz/OZtSWcol/UMgQJALesy++GdvoIDLfJX5GBQpuFgFenRiRDabxrE9MNUZ2aP
FaFp+DyAe+b4nDwuJaW2LURbr8AEZga7oQj0uYxcYw==
-----END RSA PRIVATE KEY-----
"""


def get_parser():
    parser = argparse.ArgumentParser(description='DAG trigger', add_help=True)
    parser.add_argument("-d", "--dag",  help="Dag id",                          required=True)
    parser.add_argument("-r", "--run",  help="Run id",                          required=True)
    parser.add_argument("-c", "--conf", help="Configuration, json-like string", required=True)
    parser.add_argument("-a", "--alg",  help="JWT algorithm",                   default="RS256")
    parser.add_argument("-u", "--url",  help="API url, with port number",       default="http://localhost:8080")
    return parser


def trigger_dag(dag_id, run_id, api_url, conf, alg):
    json_data = {
        "run_id": run_id,
        "conf": conf
    }
    json_data["token"] = jwt.encode(json_data, private_key, algorithm=alg).decode("utf-8")
    return requests.post(url=urljoin(api_url, f"""/api/experimental/dags/{dag_id}/dag_runs"""),
                         json=json_data)


def main(argsl=None):
    if argsl is None:
        argsl = sys.argv[1:]
    args,_ = get_parser().parse_known_args(argsl)
    res = trigger_dag(args.dag, args.run, args.url, args.conf, args.alg)
    print(res.text)


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))


#! /usr/bin/env python3

import cwltool.load_tool as load
from cwltool.context import LoadingContext
from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException
from cwltool.load_tool import (fetch_document, make_tool, resolve_and_validate_document)
from cwltool.resolver import tool_resolver
from cwltool.workflow import default_make_tool


def flatten(input_list):
    result = []
    for i in input_list:
        if isinstance(i, list):
            result.extend(flatten(i))
        else:
            result.append(i)
    return result


def conf_get_default(section, key, default):
    try:
        return conf.get(section, key)
    except AirflowConfigException:
        return default


def shortname(n):
    return n.split("#")[-1]


def load_tool(argsworkflow, loadingContext):
    loadingContext, workflowobj, uri = fetch_document(argsworkflow, loadingContext)
    loadingContext, uri = resolve_and_validate_document(loadingContext, workflowobj, uri, skip_schemas=True)
    return make_tool(uri, loadingContext)


def load_cwl(cwl_file, default_args):
    load.loaders = {}
    loading_context = LoadingContext(default_args)
    loading_context.construct_tool_object = default_make_tool
    loading_context.resolver = tool_resolver
    tool = load_tool(cwl_file, loading_context)
    it_is_workflow = tool.tool["class"] == "Workflow"
    return tool, it_is_workflow



import jwt
import logging
import requests
from airflow.models import Variable
from airflow.utils.state import State
from airflow.hooks.http_hook import HttpHook


logger = logging.getLogger(__name__)


CONN_ID = "process_report"
ROUTES = {
    "progress": "progress",
    "results":  "results",
    "status":   "status"
}
PRIVATE_KEY = "process_report_private_key"
ALGORITHM = "process_report_algorithm"


def prepare_connection(conn_id, route=""):
    http_hook = HttpHook(http_conn_id=conn_id)
    session = http_hook.get_conn()
    url = http_hook.base_url.rstrip("/") + '/' + session.headers["endpoint"].strip("/") + "/" + route.lstrip("/")
    return http_hook, session, url


def sign_with_jwt(data, private_key=None, algorithm=None):
    try:
        data = jwt.encode(payload=data,
                          key=private_key or Variable.get(PRIVATE_KEY),
                          algorithm=algorithm or Variable.get(ALGORITHM)).decode("utf-8")
    except Exception as e:
        logger.warning("Failed to sign data with JWT key: {}".format(e))
    return data


def post_progress(context, from_task=False):
    try:
        http_hook, session, url = prepare_connection(CONN_ID, ROUTES["progress"])
        dag_run = context["dag_run"]
        len_tis = len(dag_run.get_task_instances())
        len_tis_success = len(dag_run.get_task_instances(state=State.SUCCESS)) + int(from_task)
        data = sign_with_jwt(data={"state":    dag_run.state,
                                   "dag_id":   dag_run.dag_id,
                                   "run_id":   dag_run.run_id,
                                   "progress": int(len_tis_success / len_tis * 100),
                                   "error":    context["reason"] if dag_run.state == State.FAILED else ""})
        prepped_request = session.prepare_request(requests.Request("POST", url, json={"payload": data}))
        http_hook.run_and_check(session, prepped_request, {})
    except Exception as e:
        logger.warning("Failed to POST progress updates: {}".format(e))


def post_results(context):
    try:
        http_hook, session, url = prepare_connection(CONN_ID, ROUTES["results"])
        dag_run = context["dag_run"]
        results = ""
        if context["ti"].xcom_pull(task_ids="CWLJobGatherer") and len(context["ti"].xcom_pull(task_ids="CWLJobGatherer")) > 0:
            results = context["ti"].xcom_pull(task_ids="CWLJobGatherer")[0]
        data = sign_with_jwt(data={"dag_id":  dag_run.dag_id,
                                   "run_id":  dag_run.run_id,
                                   "results": results})
        prepped_request = session.prepare_request(requests.Request("POST", url, json={"payload": data}))
        http_hook.run_and_check(session, prepped_request, {})
    except Exception as e:
        logger.warning("Failed to POST results: {}".format(e))


def post_status(context):
    try:
        http_hook, session, url = prepare_connection(CONN_ID, ROUTES["status"])
        dag_run = context["dag_run"]
        ti = context["ti"]
        data = sign_with_jwt(data={"state":    ti.state,
                                   "dag_id":   dag_run.dag_id,
                                   "run_id":   dag_run.run_id,
                                   "task_id":  ti.task_id})
        prepped_request = session.prepare_request(requests.Request("POST", url, json={"payload": data}))
        http_hook.run_and_check(session, prepped_request, {})
    except Exception as e:
        logger.warning("Failed to POST status updates: {}".format(e))


def task_on_success(context):
    post_progress(context, True)
    post_status(context)


def task_on_failure(context):
    post_status(context)


def task_on_retry(context):
    post_status(context)


def dag_on_success(context):
    post_progress(context)
    post_results(context)


def dag_on_failure(context):
    post_progress(context)




import jwt
from functools import wraps
from flask import request, Response
from airflow.models import Variable


client_auth = None


def init_app(app):
    pass


def requires_authentication(function):
    @wraps(function)
    def decorated(*args, **kwargs):
        PUBLIC_KEY = "jwt_backend_public_key"
        ALGORITHM = "jwt_backend_algorithm"
        try:
            json_data = {k: v for k, v in request.get_json(force=True).copy().items() if k != "token"}
            decoded_token = jwt.decode(request.get_json(force=True)["token"], Variable.get(PUBLIC_KEY), algorithms=Variable.get(ALGORITHM))
            assert (json_data == decoded_token)
        except Exception:
            return Response("Failed to verify data", 403)
        return function(*args, **kwargs)
    return decorated



#! /usr/bin/env python3

import logging
import json
from ruamel import yaml
from six.moves import urllib
from cwltool.argparser import get_default_args
from airflow.models import DAG
from airflow.operators import BaseOperator
from airflow.exceptions import AirflowException
from airflow.utils.dates import days_ago
from .cwlstepoperator import CWLStepOperator
from .cwlutils import conf_get_default
from cwl_airflow.utils.notifier import dag_on_success, dag_on_failure
# logging.getLogger('cwltool').setLevel(conf_get_default('core', 'logging_level', 'ERROR').upper())
# logging.getLogger('salad').setLevel(conf_get_default('core', 'logging_level', 'ERROR').upper())
logging.getLogger('cwltool').setLevel(logging.ERROR)
logging.getLogger('past.translation').setLevel(logging.ERROR)
logging.getLogger('salad').setLevel(logging.ERROR)
logging.getLogger('rdflib').setLevel(logging.ERROR)

_logger = logging.getLogger(__name__)
_logger.setLevel(conf_get_default('core', 'logging_level', 'ERROR').upper())


class CWLDAG(DAG):

    def __init__(
            self,
            dag_id=None,
            cwl_workflow=None,
            default_args={},
            schedule_interval=None,
            *args, **kwargs):

        self.top_task = None
        self.bottom_task = None
        self.cwlwf = self.quick_load_cwl(cwl_workflow)

        kwargs.update({"on_failure_callback": kwargs.get("on_failure_callback", dag_on_failure),
                       "on_success_callback": kwargs.get("on_success_callback", dag_on_success)})

        init_default_args = {
            'start_date': days_ago(14),
            'email_on_failure': False,
            'email_on_retry': False,
            'end_date': None,
            'tmp_folder': conf_get_default('cwl', 'tmp_folder', '/tmp'),
            'basedir': conf_get_default('cwl', 'tmp_folder', '/tmp'),
            'no_match_user': conf_get_default('cwl', 'no_match_user', False),
            'task_retries': conf_get_default('cwl', 'retry', 1),
            'quiet': False,
            'strict': False,
            'on_error': 'continue',
            'skip_schemas': True,
            'cwl_workflow': cwl_workflow
        }

        init_default_args.update(default_args)
        merged_default_args = get_default_args()
        merged_default_args.update(init_default_args)

        super(self.__class__, self).__init__(dag_id=dag_id if dag_id else urllib.parse.urldefrag(cwl_workflow)[0].split("/")[-1].replace(".cwl", "").replace(".", "_dot_"),
                                             default_args=merged_default_args,
                                             schedule_interval=schedule_interval, *args, **kwargs)

    def quick_load_cwl(self, cwl_file):
        with open(cwl_file, "r") as input_stream:
            cwl_data = yaml.round_trip_load(input_stream, preserve_quotes=True)
        return cwl_data

    def create(self):
        if self.cwlwf["class"] in ["CommandLineTool", "ExpressionTool"]:
            cwl_task = CWLStepOperator(task_id=self.dag_id,
                                       dag=self,
                                       retries=self.default_args["task_retries"],
                                       ui_color='#5C6BC0')
        else:
            outputs = {}

            for step_id, step_val in self.cwlwf["steps"].items():
                cwl_task = CWLStepOperator(task_id=step_id,
                                           dag=self,
                                           retries=self.default_args["task_retries"],
                                           ui_color='#5C6BC0')
                outputs[step_id] = cwl_task

                for out in step_val["out"]:
                    outputs["/".join([step_id, out])] = cwl_task

            for step_id, step_val in self.cwlwf["steps"].items():
                current_task = outputs[step_id]
                if not step_val["in"]:  # need to check it, because in can be set as []
                    continue
                for inp_id, inp_val in step_val["in"].items():
                    if isinstance(inp_val, list):
                        step_input_sources = inp_val
                    elif isinstance(inp_val, str):
                        step_input_sources = [inp_val]
                    elif isinstance(inp_val, dict) and "source" in inp_val:
                        if isinstance(inp_val["source"], list):
                            step_input_sources = inp_val["source"]
                        else:
                            step_input_sources = [inp_val["source"]]
                    else:
                        step_input_sources = []

                    for source in step_input_sources:
                        parent_task = outputs.get(source, None)
                        if parent_task and parent_task not in current_task.upstream_list:
                            current_task.set_upstream(parent_task)

        # https://material.io/guidelines/style/color.html#color-color-palette
        for t in self.tasks:
            if not t.downstream_list and t.upstream_list:
                t.ui_color = '#4527A0'
            elif not t.upstream_list:
                t.ui_color = '#303F9F'

    def add(self, task, to=None):
        if not isinstance(task, BaseOperator):
            raise AirflowException(
                "Relationships can only be set between "
                "Operators; received {}".format(task.__class__.__name__))
        if to == 'top':
            self.top_task = self.top_task if self.top_task else task
            task.set_downstream([t for t in self.tasks if t.task_id != task.task_id and not t.upstream_list])
        elif to == 'bottom':
            self.bottom_task = self.bottom_task if self.bottom_task else task
            task.set_upstream([t for t in self.tasks if t.task_id != task.task_id and not t.downstream_list])

        if self.top_task and self.bottom_task:
            self.bottom_task.reader_task_id = self.top_task.task_id
            for t in self.tasks:
                if t.task_id != self.top_task.task_id:
                    t.reader_task_id = self.top_task.task_id

    def get_output_list(self):
        outputs = {}
        for out_id, out_val in self.cwlwf["outputs"].items():
            if "outputSource" in out_val:
                outputs[out_val["outputSource"]] = out_id
            else:
                outputs[out_id] = out_id
        _logger.debug("{0} get_output_list: \n{1}".format(self.dag_id, outputs))
        return outputs



import logging
import os
import io
import sys
from tempfile import mkdtemp
from json import dumps

import schema_salad.schema
from schema_salad.ref_resolver import Loader, file_uri
import ruamel.yaml as yaml

from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from ..cwlutils import load_cwl
from cwl_airflow.utils.notifier import task_on_success, task_on_failure, task_on_retry, post_status
from cwl_airflow.utils.helpers import get_folder
from cwltool.main import jobloaderctx, init_job_order

_logger = logging.getLogger(__name__)


class CWLJobDispatcher(BaseOperator):

    ui_color = '#1E88E5'
    ui_fgcolor = '#FFF'

    @apply_defaults
    def __init__(
            self,
            task_id=None,
            ui_color=None,
            tmp_folder=None,
            *args, **kwargs):
        task_id = task_id if task_id else self.__class__.__name__

        kwargs.update({"on_success_callback": kwargs.get("on_success_callback", task_on_success),
                       "on_failure_callback": kwargs.get("on_failure_callback", task_on_failure),
                       "on_retry_callback":   kwargs.get("on_retry_callback",   task_on_retry)})

        super(CWLJobDispatcher, self).__init__(task_id=task_id, *args, **kwargs)

        self.tmp_folder = tmp_folder if tmp_folder else self.dag.default_args['tmp_folder']
        if ui_color: self.ui_color = ui_color

    def cwl_dispatch(self, json):
        try:
            cwlwf, it_is_workflow = load_cwl(self.dag.default_args["cwl_workflow"], self.dag.default_args)
            cwl_context = {"outdir": mkdtemp(dir=get_folder(os.path.abspath(self.tmp_folder)), prefix="dag_tmp_")}

            _jobloaderctx = jobloaderctx.copy()
            _jobloaderctx.update(cwlwf.metadata.get("$namespaces", {}))
            loader = Loader(_jobloaderctx)

            try:
                job_order_object = yaml.round_trip_load(io.StringIO(initial_value=dumps(json)))
                job_order_object, _ = loader.resolve_all(job_order_object,
                                                         file_uri(os.getcwd()) + "/",
                                                         checklinks=False)
            except Exception as e:
                _logger.error("Job Loader: {}".format(str(e)))

            job_order_object = init_job_order(job_order_object, None, cwlwf, loader, sys.stdout)

            cwl_context['promises'] = job_order_object

            logging.info(
                '{0}: Final job: \n {1}'.format(self.task_id, dumps(cwl_context, indent=4)))

            return cwl_context

        except Exception as e:
            _logger.info(
                'Dispatch Exception {0}: \n {1} {2}'.format(self.task_id, type(e), e))
            pass
        return None

    def execute(self, context):

        post_status(context)

        _json = {}

        if 'job' in context['dag_run'].conf:
            logging.debug(
                '{0}: dag_run conf: \n {1}'.format(self.task_id, context['dag_run'].conf['job']))
            _json = context['dag_run'].conf['job']

        cwl_context = self.cwl_dispatch(_json)
        if cwl_context:
            return cwl_context
        else:
            raise Exception("No cwl context")




#! /usr/bin/env python3
import os
import subprocess
import time
from setuptools import setup, find_packages


GIT_VERSION_FILE = os.path.join('cwl_airflow','git_version')
HERE = os.path.abspath(os.path.dirname(__file__))


def get_git_tag():
    return subprocess.check_output(['git', 'describe', '--contains'], text=True).split("^")[0].strip()


def get_git_timestamp():
    gitinfo = subprocess.check_output(
        ['git', 'log', '--first-parent', '--max-count=1', '--format=format:%ct', '.'], text=True).strip()
    return time.strftime('%Y%m%d%H%M%S', time.gmtime(int(gitinfo)))


def get_description():
    README = os.path.join(HERE, 'README.md')
    with open(README, 'r') as f:
        return f.read()


def get_version():
    '''
    Tries to get pachage version with following order:
    0. default version
    1. from git_version file - when installing from pip, this is the only source to get version
    2. from tag
    3. from commit timestamp
    Updates/creates git_version file with the package version
    :return: package version
    '''
    version = '1.1.0'                                      # set default version
    try:
        with open(GIT_VERSION_FILE, 'r') as input_stream:  # try to get version info from file
            version = input_stream.read()
    except Exception:
        pass
    try:
        version = get_git_tag()                            # try to get version info from the closest tag
    except Exception:
        try:
            version = '1.1.' + get_git_timestamp()         # try to get version info from commit date
        except Exception:
            pass
    try:
        with open(GIT_VERSION_FILE, 'w') as output_stream: # save updated version to file (or the same)
            output_stream.write(version)
    except Exception:
        pass
    return version


setup(
    name='cwl-airflow',
    description='Python package to extend Airflow functionality with CWL v1.0 support',
    long_description=get_description(),
    long_description_content_type="text/markdown",
    version=get_version(),
    url='https://github.com/Barski-lab/cwl-airflow',
    download_url='https://github.com/Barski-lab/cwl-airflow',
    author='Michael Kotliar',
    author_email='misha.kotliar@gmail.com',
    license='Apache-2.0',
    include_package_data=True,
    packages=find_packages(
        exclude=[
            'docs', 'tests',
            'windows', 'macOS', 'linux',
            'iOS', 'android',
            'django'
        ]
    ),
    install_requires=[
        'cryptography',
        'cwltool == 1.0.20190621234233',
        'jsonmerge',
        'ruamel.yaml <= 0.15.77',
        'apache-airflow == 1.10.3',
        'mysqlclient >= 1.3.6, < 1.4',
        'pyjwt',
        'urllib3',
        'connexion',
        'Flask == 1.0.4'
    ],
    options={
        'app': {
            'formal_name': 'cwl-airflow',
            'bundle': 'python.org'
        },
        'macos': {
            'app_requires': [
                'toga-cocoa==0.3.0.dev11',
            ]
        },
        'linux': {
            'app_requires': [
                'toga-gtk==0.3.0.dev11',
            ]
        },
        'windows': {
            'app_requires': [
                'toga-winforms==0.3.0.dev11',
            ]
        },
        'ios': {
            'app_requires': [
                'toga-ios==0.3.0.dev11',
            ]
        },
        'android': {
            'app_requires': [
                'toga-android==0.3.0.dev11',
            ]
        },
        'django': {
            'app_requires': [
                'toga-django==0.3.0.dev11',
            ]
        },
    },
    zip_safe=False,
    scripts=["cwl_airflow/bin/cwl-airflow"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Other Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Healthcare Industry',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Operating System :: OS Independent',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 8.1',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Medical Science Apps.'
    ]
)



import os


def get_folder(abs_path, permissions=0o0775, exist_ok=True):
    try:
        os.makedirs(abs_path, mode=permissions)
    except os.error as ex:
        if not exist_ok:
            raise
    return abs_path